/** The Messenger is responsible for two-way communication
    between another process and device and the Dispatcher. */
public class Messenger extends Thread {
	
	/** The dispatcher we report to/from. */
	protected Dispatcher dispatcher;
	
	/** Time to wait between attempting to establish communication.
	    Useful as the other process or device may not be listening. */
	protected int connectionDelay;
	
	/** Are we successfully connected to the other process/device right now? */
	protected boolean connected = false;
	
	/** On a scale of 0 to 10, how much diagnostic data do we want to spew? */
	protected int verbosity = 10;
	
	/** The name of the messenger. */
	private String name;
	
	public Messenger(String name, Dispatcher dispatcher, int connectionDelay, int verbosity) {
		this.name = name;
		this.dispatcher = dispatcher;
		this.connectionDelay = connectionDelay;
		this.verbosity = verbosity;
		setDaemon(true);
		start();
	}
	
	/** Repeatedly try to connect to the device/process.  
	    Does not return unless a connection is established. */
	public void connect() { }
	
	/** Listens for messages from the other device/process, 
	    and forwards them to the dispatcher.  Does not exit
	    unless communication is broken. */
	public void receiveMessages() { }
	
	/** Returns true if the message is sent successfully (and false if there is no connection) */
	public boolean sendMessage(ScratchMessage msg, Messenger source) { return false; }
	
	/** The thread's main loop. */
	public void run() {
		while (true) {
			if (verbosity > 2)
				System.out.println(name + " is attempting to (re-)connect.");
			connect();
			connected = true;

			if (verbosity > 2)
				System.out.println(name + " is listening for messages.");				
			receiveMessages();
			connected = false;
		}
	}
	
	public boolean isConnected() {
		return connected;
	}
	
	public String name() {
		return name;
	}
	
	public String toString() {
		return name;
	}
	
	protected void forward(ScratchMessage msg) {
		dispatcher.forward(msg, this);
	}
	
	protected void forwardTo(ScratchMessage msg, String destinationName) {
		dispatcher.forwardTo(msg, this, destinationName);
	}
	
	protected void broadcastMsg(String msg) {
		forward(new ScratchMessage("broadcast \"" + msg + "\""));
	}

	// narrowcasts are used to allow the Front End and Back End to communicate with each other
	// (and not everything else)
	protected void narrowcastMsg(String msg) {
		forward(new ScratchMessage("narrowcast \"" + msg + "\""));
	}
	
	protected void updateMsg(String name, String value) {
		forward(new ScratchMessage("sensor-update " + name + " " + value));
	}
	
	protected void updateMsg(String msg) {
		forward(new ScratchMessage("sensor-update " + msg));
	}

}