ScratchMedia subclass: #SoundMedia
	instanceVariableNames: 'originalSound volume balance playingSound compressedSampleRate compressedBitsPerSample compressedData savedSound '
	classVariableNames: 'PlaceholderSound '
	poolDictionaries: ''
	category: 'Scratch-Objects'!
!SoundMedia commentStamp: '<historical>' prior: 0!
I am a Scratch sound media object. A sound can be played under script control but (obviously) it cannot be used as a costume because it is not a visual media.
!


!SoundMedia methodsFor: 'initialize' stamp: 'ee 8/12/2008 21:42'!
initialize

	super initialize.
	mediaName _ 'pop' localized.
	originalSound _ ScriptableScratchMorph popSound.
	volume _ 100.
	balance _ 50.
! !


!SoundMedia methodsFor: 'accessing' stamp: 'jm 5/4/2004 15:23'!
balance

	^ (100.0 * balance) rounded
! !

!SoundMedia methodsFor: 'accessing' stamp: 'jm 5/10/2004 18:51'!
balance: aNumber
	"Set my left-right balance, a number between 0 and 100, where 50 is centered."

	balance _ (aNumber rounded max: 0) min: 100.
	playingSound ifNotNil: [
		playingSound setPan: (balance / 100.0) volume: (volume / 100.0) for: 1].
! !

!SoundMedia methodsFor: 'accessing' stamp: 'jm 5/4/2004 15:23'!
currentSeconds

	playingSound
		ifNil: [^ 0]
		ifNotNil: [^ playingSound sounds first currentPosition].
! !

!SoundMedia methodsFor: 'accessing' stamp: 'jm 5/4/2004 15:23'!
currentSeconds: ignored
	"Currently, this doesn't do anything. It should jump to the given seconds offset."

! !

!SoundMedia methodsFor: 'accessing' stamp: 'ee 1/2/2006 14:19'!
infoString
	"Answer a string for this media, typically something about its size."

	^ self totalSeconds hhmmss
	
! !

!SoundMedia methodsFor: 'accessing' stamp: 'jm 5/10/2004 18:47'!
isSound
	"Answer true if I am a sound media object."

	^ true
! !

!SoundMedia methodsFor: 'accessing' stamp: 'ee 11/15/2007 17:21'!
loadFile: fName
	"Read my samples from a sound file."

	| snd errorString sndName mergedBuf |
	(FileDirectory default fileExists: fName)
		ifFalse: [^ self inform: 'File not found' withDetails: fName].

	errorString _ nil.
	sndName _ FileDirectory localNameFor: fName.
	Utilities informUserDuring: [:bar |
		bar value: 'Reading ', sndName, '...'.
		[snd _ SampledSound fromFileNamed: fName] ifError: [:err :rcvr |
			snd _ nil.
			errorString _ err].

		snd ifNotNil: [
			snd isStereo ifTrue: [  "merge stereo to mono"
				bar value: 'Converting ', sndName, 'to mono...'.
				mergedBuf _ snd sounds first samples.
				mergedBuf mixIn: snd sounds second samples.
				snd _ SampledSound
					samples: mergedBuf
					samplingRate: snd sounds first originalSamplingRate].

			snd originalSamplingRate > 22050 ifTrue: [  "downsample to save space"
				bar value: 'Downsampling ', sndName, '...'.
				snd _ SampledSound
					samples: (snd samples downSampledLowPassFiltering: false)
					samplingRate: snd originalSamplingRate / 2]]].

	errorString ifNotNil: [^ self error: errorString]. 
	self sound: snd.
! !

!SoundMedia methodsFor: 'accessing' stamp: 'jm 10/20/2007 19:42'!
mediaSizeInKilobytes
	"Returns the storage size of this sound in kilobytes, rounded to the nearest kilobyte."

	| bytes |
	bytes _ 0.
	compressedData
		ifNil: [
			(originalSound isKindOf: MixedSound)
				ifTrue: [
					originalSound sounds do: [:snd |
						bytes _ bytes + (2 * snd samples monoSampleCount)]]
				ifFalse: [
					bytes _ 2 * originalSound samples monoSampleCount]]
		ifNotNil: [bytes _ compressedData size].

	^ (bytes + 512) // 1024
! !

!SoundMedia methodsFor: 'accessing' stamp: 'jm 5/13/2004 17:17'!
mediaType
	"Answer a string describing this type of media."

	^ 'sound'
! !

!SoundMedia methodsFor: 'accessing' stamp: 'jm 10/17/2007 13:38'!
savedSound

	^ savedSound
! !

!SoundMedia methodsFor: 'accessing' stamp: 'jm 5/4/2004 15:23'!
sound

	^ originalSound
! !

!SoundMedia methodsFor: 'accessing' stamp: 'jm 10/17/2007 13:45'!
sound: aSampledSound
	"Set my sound."

	self pausePlaying.
	playingSound _ nil.
	originalSound _ (SampledSound
		samples: aSampledSound samples
		samplingRate: aSampledSound originalSamplingRate).
	compressedSampleRate _ compressedBitsPerSample _ compressedData _ nil.
	savedSound _ nil.
! !

!SoundMedia methodsFor: 'accessing' stamp: 'jm 5/10/2004 18:57'!
volume

	^ volume
! !

!SoundMedia methodsFor: 'accessing' stamp: 'jm 5/10/2004 18:52'!
volume: aNumber
	"Set my volume, a number between 0 and 100."

	volume _ (aNumber rounded max: 0) min: 100.
	playingSound ifNotNil: [
		playingSound setPan: (balance / 100.0) volume: (volume / 100.0) for: 1].
! !


!SoundMedia methodsFor: 'scratch ops' stamp: 'CWB 8/15/2012 12:49'!
exportFilename: fileName atBitDepth: bitDepth withMaxSampleRate: maxSampleRate for: stageOrSprite
	"Export my sound to a file with the given name."

	| fName f samples sampleRate newSound |
	fName _ fileName.
	(fName asLowercase endsWith: '.wav') ifFalse: [fName _ fName, '.wav'].
	f _ StandardFileStream newScratchFileNamed: fName.
	f ifNil: [^ self].  "could not create or open file for writing"

	samples _ originalSound samples.
	sampleRate _ originalSound originalSamplingRate.
	[sampleRate > maxSampleRate] whileTrue: [
		samples _ samples downSampledLowPassFiltering: false.
		sampleRate _ sampleRate // 2].

	newSound _ SampledSound samples: samples samplingRate: sampleRate.

	f binary.
	(bitDepth == 8) ifTrue: [
		newSound store8BitWAVSamplesAtMyRateOn: f.
	] ifFalse: [
		"newSound storeWAVSamplesOn: f."
		self error: 'exportFilename:atBitDepth:withMaxSampleRate:for: does not yet support anything other than 8 bits per sample.'
	].
	f close.
! !

!SoundMedia methodsFor: 'scratch ops' stamp: 'jm 3/20/2007 10:50'!
exportFilename: fileName for: stageOrSprite
	"Export my sound to a file with the given name."

	| fName f |
	fName _ fileName.
	(fName asLowercase endsWith: '.wav') ifFalse: [fName _ fName, '.wav'].
	f _ StandardFileStream newScratchFileNamed: fName.
	f ifNil: [^ self].  "could not create or open file for writing"
	f binary.
	originalSound storeWAVSamplesOn: f.
	f close.
! !

!SoundMedia methodsFor: 'scratch ops' stamp: 'jm 5/8/2005 11:37'!
isPlaying
	"Answer true if my sound is playing."

	^ (playingSound notNil and: [playingSound isPlaying])
! !

!SoundMedia methodsFor: 'scratch ops' stamp: 'jm 5/4/2004 15:23'!
pausePlaying
	"Pauses sound playback."

	playingSound ifNotNil: [playingSound pause].
! !

!SoundMedia methodsFor: 'scratch ops' stamp: 'jm 5/4/2004 15:23'!
playFrom: secs
	"Play from the given point through the end of the sound."

	self playFrom: secs to: self totalSeconds.
! !

!SoundMedia methodsFor: 'scratch ops' stamp: 'jm 5/10/2004 18:58'!
playFrom: startSecs to: endSecs
	"Play the given portion of the sound. The start and end offsets are in seconds at the original sound's sampling rate regardless of the current playback rate."

	| totalSecs sampleCount startIndex endIndex snd |
	self pausePlaying.
	originalSound ifNil: [^ self].
	totalSecs _ self totalSeconds.
	sampleCount _ originalSound samples monoSampleCount.
	startIndex _ (startSecs * sampleCount // totalSecs) max: 1.
	endIndex _ (endSecs * sampleCount // totalSecs) min: sampleCount.
	startIndex < endIndex ifFalse: [^ self].

	playingSound _ MixedSound new.
	snd _ (SampledSound new
		setSamples: originalSound samples
		from: startIndex to: endIndex
		samplingRate: originalSound originalSamplingRate
		loudness: 1.0).
	playingSound add: snd pan: (balance / 100.0) volume: (volume / 100.0).
	playingSound play.
! !

!SoundMedia methodsFor: 'scratch ops' stamp: 'jm 5/4/2004 15:23'!
playFromStart
	"Play my sound from beginning to end."

	self playFrom: 0 to: self totalSeconds.
! !

!SoundMedia methodsFor: 'scratch ops' stamp: 'jm 12/17/2004 13:27'!
rewindSound
	"Rewind the playing sound, if any."

	playingSound ifNotNil: [playingSound sounds first reset].
! !

!SoundMedia methodsFor: 'scratch ops' stamp: 'jm 5/13/2004 17:12'!
startPlaying
	"Resume playing if the sound was paused before the end. Otherwise, start playing from the beginning."

	(playingSound isNil or:
	 [playingSound isEmpty or:
	 [playingSound samplesRemaining = 0]])
		ifTrue: [self playFromStart]
		ifFalse: [playingSound resumePlaying].
! !

!SoundMedia methodsFor: 'scratch ops' stamp: 'jm 4/23/2008 09:46'!
thumbnailFormExtent: extent
	"Answer a Form of the given extent containing a thumbnail of me containing the sound icon."

	| offForm onForm f icon |
	offForm _ ScratchFrameMorph skinAt: #scratchSound.
	onForm _ ScratchFrameMorph skinAt: #scratchSoundOn.
	f _ (Form extent: extent depth: 8).
	playingSound
		ifNil: [icon _ offForm]
		ifNotNil: [icon _ self isPlaying ifTrue: [onForm] ifFalse: [offForm]].
	icon displayOn: f at: (f extent - icon extent) // 2 rule: Form paint.
	^ f
! !

!SoundMedia methodsFor: 'scratch ops' stamp: 'jm 5/4/2004 15:23'!
totalSeconds

	originalSound
		ifNil: [^ 0]
		ifNotNil: [^ originalSound samples size asFloat / originalSound originalSamplingRate].
! !


!SoundMedia methodsFor: 'compressing' stamp: 'jm 10/17/2007 13:34'!
compressBitsPerSample: bits saveOriginal: saveFlag
	"Compress my sound if possible. Return true if compression is done."

	| samples |
	self isBuiltInSound ifTrue: [^ false].  "don't compress built-in sounds"
	compressedData ifNotNil: [^ false].  "don't recompress"
	((originalSound isKindOf: SampledSound) not or:
	 [originalSound isStereo]) ifTrue: [^ false].

	saveFlag
		ifTrue: [savedSound _ originalSound]
		ifFalse: [savedSound _ nil].

	samples _ originalSound samples.
	compressedSampleRate _ originalSound originalSamplingRate.
	[compressedSampleRate >= 32000] whileTrue: [
		samples _ samples downSampledLowPassFiltering: false.
		compressedSampleRate _ compressedSampleRate // 2].

	compressedBitsPerSample _ bits.
	compressedData _ (ADPCMCodec newBitsPerSample: bits) encodeSoundBuffer: samples.

	^ true

! !

!SoundMedia methodsFor: 'compressing' stamp: 'jm 10/12/2007 20:56'!
compressedBitsPerSample

	^ compressedBitsPerSample
! !

!SoundMedia methodsFor: 'compressing' stamp: 'jm 10/12/2007 20:57'!
compressedData

	^ compressedData
! !

!SoundMedia methodsFor: 'compressing' stamp: 'jm 10/12/2007 20:56'!
compressedSampleRate

	^ compressedSampleRate
! !

!SoundMedia methodsFor: 'compressing' stamp: 'jm 10/17/2007 13:34'!
decompress
	"Decompress sound from compressed data so result of compression can be heard. Do nothing if sound is not compressed."

	| samples |
	compressedData ifNil: [^ self].

	samples _ (ADPCMCodec newBitsPerSample: compressedBitsPerSample)
		decodeCompressedData: compressedData.
	originalSound _ SampledSound
		samples: samples
		samplingRate: compressedSampleRate.

! !

!SoundMedia methodsFor: 'compressing' stamp: 'jm 8/1/2007 17:51'!
isBuiltInSound
	"Answer true if my sound is one of the Scratch builtin sounds."

	((mediaName = 'pop') and:
	 [originalSound samples size = ScriptableScratchMorph popSound samples size])
		ifTrue: [^ true].
	((mediaName = 'meow') and:
	 [originalSound samples size = ScriptableScratchMorph meowSound samples size])
		ifTrue: [^ true].

	^ false
! !

!SoundMedia methodsFor: 'compressing' stamp: 'jm 8/1/2007 17:51'!
reduceSamplingRate
	"Reduce my sampling rate to under 22 ksamples/sec (typically to 11.025 ksamples/sec). Answer true if the sampling rate was reduced."

	originalSound originalSamplingRate < 22000 ifTrue: [^ false].  "already reduced"
	self isBuiltInSound ifTrue: [^ false].  "built-in sound"

	self pausePlaying.
	playingSound _ nil.

	[originalSound originalSamplingRate < 22000] whileFalse: [
		originalSound _ SampledSound
			samples: (originalSound samples downSampledLowPassFiltering: true)
			samplingRate: originalSound originalSamplingRate // 2].

	^ true
! !

!SoundMedia methodsFor: 'compressing' stamp: 'jm 10/17/2007 13:08'!
revertToUncompressed
	"Revert to my original, uncompressed sound after compression with the 'saveOriginal' flag."

	savedSound ifNotNil: [
		originalSound _ savedSound.
		compressedSampleRate _ nil.
		compressedBitsPerSample _ nil.
		compressedData _ nil.
		savedSound _ nil].
! !

!SoundMedia methodsFor: 'compressing' stamp: 'jm 10/17/2007 13:38'!
shareSoundWith: aSoundMedia
	"Share my sound and compressed sound with the given SoundMedia."

	self sound: aSoundMedia sound.
	compressedSampleRate _ aSoundMedia compressedSampleRate.
	compressedBitsPerSample _ aSoundMedia compressedBitsPerSample.
	compressedData _ aSoundMedia compressedData.
	savedSound _ aSoundMedia savedSound.
! !


!SoundMedia methodsFor: 'copying' stamp: 'jm 5/4/2004 15:23'!
updateReferencesUsing: aDictionary
	"Nil out my playing sound."

	super updateReferencesUsing: aDictionary.
	playingSound _ nil.
! !


!SoundMedia methodsFor: 'object i/o' stamp: 'jm 8/1/2007 17:24'!
fieldsVersion

	^ 2
! !

!SoundMedia methodsFor: 'object i/o' stamp: 'jm 10/17/2007 13:17'!
initFieldsFrom: anObjStream version: classVersion

	| samples |
	super initFieldsFrom: anObjStream version: classVersion.
	self initFieldsNamed: #(
		originalSound
		volume
		balance
	) from: anObjStream.

	classVersion = 1 ifTrue: [^ self].

	"fields added in version 2"
	self initFieldsNamed: #(
		compressedSampleRate
		compressedBitsPerSample
		compressedData
	) from: anObjStream.

	compressedData ifNotNil: [
		samples _ (ADPCMCodec newBitsPerSample: compressedBitsPerSample)
			decodeCompressedData: compressedData.

		originalSound _ SampledSound
			samples: samples
			samplingRate: compressedSampleRate].

! !

!SoundMedia methodsFor: 'object i/o' stamp: 'jm 10/17/2007 13:02'!
storeFieldsOn: anObjStream

	| oldSnd |
	super storeFieldsOn: anObjStream.

	oldSnd _ originalSound.
	compressedData ifNotNil: [originalSound _ PlaceholderSound].  "for backward compatability"

	self storeFieldsNamed: #(
		originalSound
		volume
		balance
		compressedSampleRate
		compressedBitsPerSample
		compressedData
	) on: anObjStream.

	originalSound _ oldSnd.
! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SoundMedia class
	instanceVariableNames: ''!

!SoundMedia class methodsFor: 'class initialization' stamp: 'jm 10/17/2007 13:00'!
initialize
	"self initialize"

	PlaceholderSound _ SampledSound samples: SoundBuffer new samplingRate: 22050.
! !


SoundMedia initialize!
